

# Cell
import open3d as o3d
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
import yaml
import math
from queue import Queue, PriorityQueue
plt.rcParams["figure.figsize"] = (16, 16)

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Cell
import yaml
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
import math

# Cell
import open3d as o3d
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap, LinearSegmentedColormap
import pylab as plot
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import yaml
import math
import os
from queue import Queue, PriorityQueue
import random
plt.rcParams["figure.figsize"] = (16, 16)